; Fargo IDE header
;=============================================================================
; Last save: Fri Nov 12 19:02:10  1999
; 278.17 1.0 1.0 1.0 1.0 1.0
;=============================================================================
/*******************************************************************************
 * Header file for FORTH92
 *------------------------------------------------------------------------------
Copyright (c) 1999 David Kuehling  <dvdkhlng TA gmx TOD de>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice (including the next
paragraph) shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
 *******************************************************************************/

/******************************************************************************
 * FORTH92's registers
 */
#define		PSB	d4/*   parameter-stack-base (bottom)*/
#define		RST	d5/*   return-stack-top */
#define	 	FSB	d6/*   float-stack-base (bottom) */
#define		WP	d7/*   word-pointer */
#define		WP.w	WP|.w
#define		IP	a2/*   instruction-pointer */
#define		BP	a3/*   base-pointer (for relative addresses) */
#define		NP	a4/*   next-pointer */
#define		FSP	a5/*   return-stack-pointer */
#define		RSP	a6/*   return-stack-pointer */
#define		PSP	a7/*   parameter-stack-pointer */


/******************************************************************************
 * Memory allocation and flag constants
 */
	// amount of free memory in dictionary, for user-defined words (bytes)
#define	USER_SPACE 16768
	// amount of free memory, that has to be at the end of the dictionary ( for <#, PAD etc. )
#define	DICT_END_SPACE 256
	// number of return addresses, the returnstack can hold
#define RS_SIZE 256
	// number of bytes, the returnstack can overflow (this value must be greater or equal
	// than 20 since this area is used as temporary buffer)
#define	RS_MAX_OVERFLOW 20
	// number of bytes, the returnstack can underflow (this value must be greater or equal
	// than 20 since this area is used as temporary buffer)
#define	RS_MAX_UNDERFLOW 20
	// maximum number of bytes, the parameterstack can hold (since the calculator's internal
	// stack is used, this value is only used for checking, it doesn't controll the 
	// allocation of any memory)
#define PS_SIZE 8192
	// maximum number of floating point numbers, the float stack can hold
#define FS_SIZE 32
	// maximum number of wordlists
#define MAX_WORDLISTS 8
	// maximum number of wordlists in search order
#define MAX_ORDER 8
	// size of terminal input buffer (characters)
#define	TIB_SIZE 84
	// size of file input buffer (characters)
#define	FIB_SIZE 128
	// size of region addressed by PAD (characters)
#define PAD_SIZE 84
	// maximum stack-underflow - size of unused safety-area in Parameter Stack (PSP, a7)
#define	MAX_UNDERFLOW 128
	// Create centralized address interpreter (influences the macro DO_NEXT)
	// Note that there are still some commands that contain Next instead of jumping to it 
	// (e.g. DoCall)

/******************************************************************************
 * Error numbers
 */
#define ERR_S_OVERFLOW		  -3/* stack overflow				 	*/
#define ERR_S_UNDERFLOW		  -4/* stack underflow				 	*/
#define ERR_RS_OVERFLOW		  -5/* return stack overflow			 	*/
#define ERR_RS_UNDERFLOW	  -6/* return stack underflow			 	*/
#define ERR_DICT_OVERFLOW 	  -8/* dictionary overflow			 	*/
#define ERR_INV_ADDRESS		  -9/* invalid memory address			 	*/
#define ERR_DIVISION		 -10/* division by zero				 	*/
#define ERR_UNDEF_WORD		 -13/* undefined word				 	*/
#define ERR_ZERO_STR		 -16/* attempt to use zero length string as name 	*/
#define ERR_PNOS_OVERFLOW	 -17/* pictured numeric output string overflow		*/
#define ERR_DEF_NAME_TOO_LONG	 -19/* definition name too long				*/
#define ERR_CTRL_STRUCT_MISMATCH -22/* control structure mismatch			*/
#define ERR_ALIGN		 -23/* address alignment exception			*/
#define ERR_USER_INTERRUPT	 -28/* user interrupt					*/
#define ERR_INV_NAME_ARG	 -32/* invalid name argument (e.g., TO xxx)		*/
#define ERR_BLK_READ		 -33/* block read exception				*/
#define ERR_BLK_WRITE		 -34/* block write exception				*/
#define ERR_INV_BLK		 -35/* invalid block number				*/
#define ERR_INV_FILE_POS	 -36/* invalid file position				*/
#define ERR_FILE_IO		 -37/* file I/O exception				*/
#define ERR_NON_EXISTENT_FILE	 -38/* non-existent file				*/
#define ERR_FS_OVERFLOW		 -44/* floating-point stack overflow			*/
#define ERR_FS_UNDERFLOW	 -45/* floating-point stack underflow			*/
#define ERR_SO_OVERFLOW		 -49/* search-order overflow				*/
#define ERR_SO_UNDERFLOW	 -50/* search-order underflow				*/
#define ERR_ILLEG_INSTR		-256/* illegal machine code instruction			*/
#define	ERR_RESTORE_INPUT	-257/* RESTORE-INPUT error				*/

/******************************************************************************
 * The address interpreter NEXT
 */
#define CENTRAL_NEXT

/** Threaded-code interpreter */
#define NEXT {
		move.w	(IP)+, WP
		move.w	0(BP,WP.w), d0
		jmp	0(BP,d0.w)
}
/** create centralized or decentralized threaded code interpreter */
#define	DO_NEXT {
	#ifdef CENTRAL_NEXT
		jmp	(NP)
	#else
		NEXT
	#endif
}

/******************************************************************************
 * Macros for creating FORTH code
 */
	// add names to internal routines -- hepful for decompiling but increases FORTH92's size
//#define	NAMED_INTERNALS

/** create a relative pointer ( accessed by XXXX(BP) ) */
#define	RELP(ptr) (ptr)-RelativePointerBase

/** create an execution-token */
#define	XT(ptr) {
		dc.w	RELP(ptr)
}

/** create a literal */
#define LIT(num) {
		XT(lit)
		dc.w	num
}

/** create a double literal */
#define 2LIT(num) {
		XT(two_lit)
		dc.l	num
}

/** create a float literal */
#define FLIT(magnitude,exponent) {
		XT(flit)
		dc.l	magnitude
		dc.w	exponent
}

/** create an SLITERAL string */
#define	SLIT(str...) {
	#local \start,\end
		XT(slit)			// execution token of string-literal-handler
		dc.b	\end-\start		// length of string
\start:		dc.b	str			// string
\end:		ds.b	0
		EVEN				// continue at even address
	#endlocal
}

/** create a dot-quote-string */
#define	STR(str...) {
	#local \start,\end
		XT(dot_quote)			// execution token of dot-string-handler
		dc.b	\end-\start		// length of string
\start:		dc.b	str			// string
\end:		ds.b	0
		EVEN				// continue at even address
	#endlocal
}
/** create a TO-access */
#define TO(value) {
		XT(value+4)
}

/** 
 * Create a word header 
 *----------------------
 *  Format      +------------------------------------+
 *              |  Link Field (2 bytes)              | ----> points to BitField of previous word
 *              +------------------------------------+
 *              |  BitField: 7  6  5  4 3 2 1 0      | <-- S: Smudge bit  I: Immediate bit  L: Name length
 *              |            1  S  I  L L L L L      |        (bit 7 is set, to indicate the name's begin)
 *              |------------------------------------|
 *              |  Name ... (not NUL terminated)     |
 *              |  evtl. dummy byte, to align        |
 *              |  remaining words on even address   |
 *              +------------------------------------+
 *              |  Code Field (2 bytes)              |
 *              +------------------------------------+
 *              |  body (may have any even length)   |
 *              +------------------------------------+
 */

			// link-field of first word points to NULL
#define LINK_PTR RelativePointerBase

#define WORD_HEADER(lbl,name,flags...) {
		dc.w	RELP(LINK_PTR)			// link field
__|lbl:	#ifdef flags
		dc.b	\nameEnd-\nameStart+flags+128	// Bitfield (length of name + IMMEDATE bit + SMUDGE bit)
	#else
		dc.b	\nameEnd-\nameStart+128		// Bitfield (length of name + IMMEDATE bit + SMUDGE bit)
	#endif

\nameStart:	dc.b	name				// name field
\nameEnd:	ds.b	0
		EVEN
lbl:
	#equ LINK_PTR __|lbl
}

/** Create a word header for an internal word. An internal word won't be linked to the 
 ** dictionary and only gets a name if the macro NAMED_INTERNALS is defined. */
#define INTERNAL_WORD_HEADER(lbl,name) {
	#ifdef	NAMED_INTERNALS
		dc.b	\nameEnd-\nameStart+128
\nameStart:	dc.b	name
\nameEnd:	ds.b	0
	#endif
		EVEN
lbl:
}

/** bit numbers and masks for BitField of entry  **/
#define	SMUDGE_BIT 6
#define IMMEDIATE_BIT 5
#define SMUDGE 64
#define	IMMEDIATE 32


/** create a primary definition */
#define PRIM_WORD(lbl,name,flags...) {
	#ifdef flags			
		WORD_HEADER(lbl,name,flags)
	#else
		WORD_HEADER(lbl,name)
	#endif
		XT(lbl+2)				// code field
}

#define NOOP_WORD(lbl,name,flags...) {
	#ifdef flags			
		WORD_HEADER(lbl,name,flags)
	#else
		WORD_HEADER(lbl,name)
	#endif
		XT(Next)
}

/** create a secundary definition */
#define SEC_WORD(lbl,name,flags...) {
	#ifdef flags			
		WORD_HEADER(lbl,name,flags)
	#else
		WORD_HEADER(lbl,name)
	#endif
		XT(DoCall)				// code field
}

/** create a constant */
#define CONSTANT(lbl,name,value) {
		WORD_HEADER(lbl,name)
		XT(DoCon)				// code field: CONSTANT-handler
		dc.w	value
	#equ LINK_PTR __|lbl
}

/** create a double constant */
#define 2CONSTANT(lbl,name,value) {
		WORD_HEADER(lbl,name)
		XT(Do2Con)				// code field: 2CONSTANT-handler
		dc.l	value
}

/** create a variable */
#define VARIABLE(lbl,name,value) {
		WORD_HEADER(lbl,name)
		XT(DoVar)				// code field: VARIABLE-handler
		dc.w	value
}

/** create a float constant */
#define FCONSTANT(lbl,name,significant,exponent) {
		WORD_HEADER(lbl,name)
		XT(DoFCon)				// code field: FCONSTANT-handler
		dc.l	significant
		dc.w	exponent
}

/** create a value */
#define VALUE(lbl,name,value) {
		CONSTANT(lbl,name,value)
		XT(DoTo)				// 2nd code field: TO-handler
}

/** create a word that does the same as a special primary definition */
#define SAME_PRIM_WORD(lbl,name,same,flags...) {
	#ifdef flags			
		WORD_HEADER(lbl,name,flags)
	#else
		WORD_HEADER(lbl,name)
	#endif
		XT(same+2)				// code field
}

/** create an internal primary word */
#define INTERNAL_PRIM_WORD(lbl,name) {
		INTERNAL_WORD_HEADER(lbl,name)
		XT(lbl+2)
}

/** create an internal secundary word */
#define INTERNAL_SEC_WORD(lbl,name) {
		INTERNAL_WORD_HEADER(lbl,name)
		XT(DoCall)
}

/** create an internal VARIABLE */
#define INTERNAL_VAR(lbl,name,value) {
		INTERNAL_WORD_HEADER(lbl,name)
		XT(DoVar)
		dc.w	value
}

/** create an internal CONSTANT */
#define INTERNAL_CONST(lbl,name,value) {
		INTERNAL_WORD_HEADER(lbl,name)
		XT(DoCon)
		dc.w	value
}

/** create an internal FCONSTANT */
#define INTERNAL_FCONST(lbl,name,significant,exponent) {
		INTERNAL_WORD_HEADER(lbl,name)
		XT(DoFCon)
		dc.l	significant
		dc.w	exponent
}

/** create an internal VALUE */
#define INTERNAL_VALUE(lbl,name,value) {
	INTERNAL_CONST(lbl,name,value)
	XT(DoTo)
}

/*******************************************************************************
 * Create Control Structures
 *******************************************************************************/

/** create a label for a branch-command */
#define LBL(lbl) {
		dc.w	RELP(lbl)
}

/** create a IF ... [ ELSE ... ] THEN structure */
#define IF(IFbody,ELSEbody...) {
	#local \then, \else
	#ifndef ELSEbody
		XT(question_bra) LBL(\then)		// IF
		IFbody					// ...
	#else						// THEN
		XT(question_bra) LBL(\else)		// IF
		IFbody					// ...
		XT(_bra) LBL(\then)			// ELSE
\else:		ELSEbody				// ...
	#endif						// THEN
\then:	#endlocal
}

/** create a BEGIN...AGAIN structure */
#define BEGIN_AGAIN(body) {
	#local \begin
\begin:		body					// BEGIN ...
		XT(bra_) LBL(\begin)			// AGAIN
	#endlocal
}

/** create a BEGIN...UNTIL structure */
#define BEGIN_UNTIL(body) {
	#local \begin
\begin:		body					// BEGIN ...
		XT(question_bra_) LBL(\begin)		// UNTIL
	#endlocal
}

/** create a BEGIN...WHILE...REPEAT structure */
#define BEGIN_WHILE(body1,body2) {
	#local \begin, \end
\begin:		body1					// BEGIN ...
		XT(question_bra) LBL(\end)		// WHILE
		body2					// ...
		XT(bra_) LBL(\begin)			// REPEAT
\end:	#endlocal
}


/** create a DO...LOOP structure */
#define DO_LOOP(body) {
	#local \begin, \end
		XT(do) LBL(\end)			// DO
\begin:		body					// ...
		XT(loop) LBL(\begin)			// LOOP
\end:	#endlocal
}

/** create a ?DO...LOOP structure */
#define QDO_LOOP(body) {
	#local \begin, \end
		XT(question_do) LBL(\end)		// ?DO
\begin:		body					// ...
		XT(loop) LBL(\begin)			// LOOP
\end:	#endlocal
}

/** create a DO...+LOOP structure */
#define DO_PLOOP(body) {
	#local \begin, \end
		XT(do) LBL(\end)			// DO
\begin:		body					// ...
		XT(plus_loop) LBL(\begin)		// +LOOP
\end:	#endlocal
}

/** create a ?DO...+LOOP structure */
#define QDO_PLOOP(body) {
	#local \begin, \end
		XT(question_do) LBL(\end)		// ?DO
\begin:		body					// ...
		XT(plus_loop) LBL(\begin)		// +LOOP
\end:	#endlocal
}

/** create a CASE (... OF ... ENDOF )* ... ENDCASE structure */
#define CASE(beforeOF,behindOF,default,more...) {
	#local \next, \end
		beforeOF				// ...
		XT(of) LBL(\next)			// OF
		behindOF				// ...
		XT(bra_) LBL(\end)			// ENDOF
\next:	#ifdef more
		CASE(default,more)			// [ OF ... ENDOF ... ]
	#else
		default					// ...
		XT(DROP)				// [ ENDCASE ]
	#endif
\end:	#endlocal
}



































































